
/* Terminality - a portable terminal handling library
 * Copyright (C) 1998-2002, Emil Mikulic.
 * This is LGPL - look at COPYING.LIB
 */

/* Project:     Terminality/GUI
 * File:	checkbox.cpp
 * Author:      Michal Safranek
 * Description: Implements the checkbox class
 */

#include <gui.h>

const char checkbox_rcsid[] =
"$Id: checkbox.cpp,v 1.9 2002/07/26 01:39:40 darkmoon Exp $";



/* Checkbox elements */
void checkbox::element(const checkbox_element el)
{
	if (has_color()) {
		/* Pretty colors */
		switch (el) {
		case CB_left:
			if (focus)
				setcolor(SELECTED_BOX,SELECTED_BOX_BG);
			else
				setcolor(BOX,BOX_BG);
			printw("%c", LBOX_CHR);
			break;

		case CB_check:
			if (focus)
				setcolor(SELECTED_CHECK,
						SELECTED_CHECK_BG);
			else
				setcolor(CHECK,CHECK_BG);
			if (selected)
				printw("%c", CHECK_CHECKED_CHR);
			else
				printw("%c", CHECK_UNCHECKED_CHR);
			break;

		case CB_label:
			if (focus)
				setcolor(SELECTED_LABEL,
						SELECTED_LABEL_BG);
			else
				setcolor(LABEL,LABEL_BG);
			break;

		case CB_right:
			if (focus)
				setcolor(SELECTED_BOX,SELECTED_BOX_BG);
			else
				setcolor(BOX,BOX_BG);
			printw("%c", RBOX_CHR);
			break;
		}
	} else {
		/* Plain */
		switch (el) {
		case CB_left:
			if (focus)
				highvideo();
			else
				normvideo();
			printw("%c", LBOX_CHR);
			break;

		case CB_check:
			if (focus)
				normvideo();
			else
				highvideo();
			if (selected)
				printw("%c", CHECK_CHECKED_CHR);
			else
				printw("%c", CHECK_UNCHECKED_CHR);
			break;

		case CB_label:
			if (focus)
				normvideo();
			else
				lowvideo();
			break;

		case CB_right:
			if (focus)
				highvideo();
			else
				normvideo();
			printw("%c",RBOX_CHR);
			break;
		}
	}
}



#define CB_SETUP()	BACKGROUND = COLOR_CHECKBOX_BG; \
			LABEL_BG = COLOR_CHECKBOX_LABELBG; \
			CHECK_BG = COLOR_CHECKBOX_CHECKBG; \
			BOX_BG = COLOR_CHECKBOX_BOXBG; \
			LABEL = COLOR_CHECKBOX_LABEL; \
			SELECTED_LABEL_BG = COLOR_CHECKBOX_SLABELBG; \
			SELECTED_CHECK_BG = COLOR_CHECKBOX_SCHECKBG; \
			SELECTED_BOX_BG = COLOR_CHECKBOX_SBOXBG; \
			SELECTED_LABEL = COLOR_CHECKBOX_SLABEL; \
			BOX = COLOR_CHECKBOX_BOX; \
			SELECTED_CHECK = COLOR_CHECKBOX_SCHECK; \
			SELECTED_BOX = COLOR_CHECKBOX_SBOX; \
			CHECK = COLOR_CHECKBOX_CHECK; \
			LBOX_CHR = '['; RBOX_CHR = ']'; \
			CHECK_CHECKED_CHR = 'x'; \
			CHECK_UNCHECKED_CHR = ' ';



// Construct checkbox at <x,y>
checkbox::checkbox(const int xx, const int yy, bool sel)
{
	x = xx;
	y = yy;
	width = 4;
	focus = false;
	selected = sel;
	label = NULL;
	type_id = Checkbox;
	visible = true;
	fixed_colors = false;
	spc_handle = NULL;
	CB_SETUP();
}



// Construct checkbox at <x,y> using <s> for a buffer
checkbox::checkbox(const int xx, const int yy, bool sel, const char *s)
{
	x = xx;
	y = yy;
	focus = false;
	selected = sel;
	width = 4 + strlen(s);
	label = (char*)xmalloc(strlen(s)+1);
	strcpy(label, s);
	type_id = Checkbox;
	visible = true;
	fixed_colors = false;
	spc_handle = NULL;
	CB_SETUP();
}



// Construct checkbox at <x,y> using <s> for a buffer, label's width <w>
checkbox::checkbox(const int xx, const int yy, bool sel,
		const char *s, const int w)
{
	x = xx;
	y = yy;
	focus = false;
	selected = sel;
	width = 4 + w;
	label = (char*)xmalloc(w+1);
	strncpy(label, s, w);
	label[w] = 0;
	type_id = Checkbox;
	visible = true;
	fixed_colors = false;
	spc_handle = NULL;
	CB_SETUP();
}



// Destroy checkbox
checkbox::~checkbox()
{
	// Deallocate buffer!
	xfree(label);
}



// Draw checkbox
void checkbox::draw(void)
{
	char *temp;

	if(!visible) return;

	// Draw empty box
	bgcolor(BACKGROUND);
	gotoxy(x,y);
	for(int i=0; i<width; i++)
		printw(" ");

	// Go to beginning of box
	gotoxy(x,y);
	element(CB_left);
	element(CB_check);
	element(CB_right);
	element(CB_label);

	// Write out label
	temp = (char*) xmalloc((width-4)+1+1); // label size+space+null
	temp[0]=' ';
	memcpy(temp+1, label, width-4);
	temp[(width-4)+1] = 0;
	printw(temp);
	xfree(temp);
	gotoxy(x+1,y);
}



/* Non-blocking get function
 * Returns:
 *    0 - enter, 27 - ESC, -1 - up, 1 - down
 */
int checkbox::getnb(void)
{
	key c;
	int result = 101;
	cursor cs = get_cursor();
	bool old_vis = visible;

	visible = true;
	set_cursor(none);	// cursor
	focus = true;		// highlighting
	draw();			// draw textbox

	/* Get the input */
	do
	{
		/* Update screen */
		update();

		/* Get letter */
		c = readkey();
		c = keyhandler(c, Checkbox);
		c = handle_key(c, Checkbox, this);

		/* Handle input */
		switch (c)
		{
		// filter
		case KEY_ENTER: result = 0; break;
		case KEY_ESC: result = 27; break;
		case KEY_UP: result = -1; break;
		case KEY_DOWN:
		case KEY_TAB: result = 1; break;
		case ' ':
			      if (selected)
				      selected = false;
			      else
				      selected = true;
			      if(spc_handle)
				      spc_handle(selected);
			      draw();
				  if(onchange)
					  onchange(Checkbox, this);
			      break;

		case KEY_NOTHING:
			      break;

		default:
			      beep();
		}
	}
	while (result == 101);

	visible = old_vis;

	focus = false; // Kill highlighting
	set_cursor(cs);

	return result;
}



// Blocking get (cannot be aborted)
bool checkbox::get(void)
{
	while (1)
	{
		switch(getnb())
		{
		case 27: return 0;
		case 0: return selected;
		}
	}
}



// Change color scheme
void checkbox::change_scheme(void)
{
	if (fixed_colors) return;
	BACKGROUND = COLOR_CHECKBOX_BG; LABEL_BG = COLOR_CHECKBOX_LABELBG;
	CHECK_BG = COLOR_CHECKBOX_CHECKBG; BOX_BG = COLOR_CHECKBOX_BOXBG;
	LABEL = COLOR_CHECKBOX_LABEL;
	SELECTED_LABEL_BG = COLOR_CHECKBOX_SLABELBG;
	BOX = COLOR_CHECKBOX_BOX; SELECTED_CHECK = COLOR_CHECKBOX_SCHECK;
	SELECTED_BOX = COLOR_CHECKBOX_SBOX; CHECK = COLOR_CHECKBOX_CHECK;
	SELECTED_CHECK_BG = COLOR_CHECKBOX_SCHECKBG;
	SELECTED_BOX_BG = COLOR_CHECKBOX_SBOXBG;
	SELECTED_LABEL = COLOR_CHECKBOX_SLABEL;
}

