
/* Terminality - a portable terminal handling library
 * Copyright (C) 1998-2002, Emil Mikulic.
 * This is LGPL - look at COPYING.LIB
 */

/* Project:     Terminality
 * File:	extra.c
 * Authors:     Volker Loew, Emil Mikulic
 * Description: Provide some extra functions
 */

#include "extra.h"
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>

const char extra_rcsid[] =
"$Id: extra.c,v 1.33 2002/07/26 01:39:40 darkmoon Exp $";



/* Draws a box using the corner and line special chars. */
void drawbox (int x1, int y1, int x2, int y2)
{
	int i, tmp;

	/* Switch values if necessary */
	if (y1 > y2)
	{
		tmp = y2; y2 = y1; y1 = tmp;
	}
	if (x1 > x2)
	{
		tmp = x2; x2 = x1; x1 = tmp;
	}

	/* Upper left corner */
	gotoxy(x1,y1);
	writech(C_UL);

	/* Top line */
	for (i = x1+1; i<x2; i++)
		writech(HLINE);

	/* Upper right corner */
	writech(C_UR);

	/* Left and right sidelines */
	for (i = y1+1; i<y2; i++)
	{
		gotoxy(x1,i);
		writech(VLINE);
	
		gotoxy(x2,i);
		writech(VLINE);
	}

	/* Lower left corner */
	gotoxy(x1,y2);
	writech(C_LL);

	/* Bottom line */
	for (i = x1+1; i<x2; i++)
		writech(HLINE);

	/* Lower right corner */
	writech(C_LR);
}



/* Draws a box with different coloured elements to make it look "3D" */
void box_3d (int x1, int y1, int x2, int y2, color basecolor, 
	color lite, color dark)
{
	int i, j, tmp;

	/* Switch values if necessary */
	if (y1 > y2)
	{
		tmp = y2; y2 = y1; y1 = tmp;
	}
	if (x1 > x2)
	{
		tmp = x2; x2 = x1; x1 = tmp;
	}

	/* On a coloured square */
	bgcolor(basecolor);

	/* Upper left corner */
	fgcolor(lite);
	gotoxy(x1,y1);
	writech(C_UL);

	/* Top line */
	for (i = x1+1; i<x2; i++)
		writech(HLINE);

	/* Upper right corner */
	fgcolor(dark);
	writech(C_UR);

	/* Left and right sidelines */
	for (i = y1+1; i<y2; i++)
	{
		/* Left sideline */
		fgcolor(lite);
		gotoxy(x1,i);
		writech(VLINE);

		/* Fill box with bg color */
		for (j = x1; j<x2; j++)
			writech(' ');
	
		/* Right sideline */
		fgcolor(dark);
		gotoxy(x2,i);
		writech(VLINE);
	}

	/* Lower left corner */
	fgcolor(lite);
	gotoxy(x1,y2);
	writech(C_LL);

	/* Bottom line */
	fgcolor(dark);
	for (i = x1+1; i<x2; i++)
		writech(HLINE);

	/* Lower right corner */
	writech(C_LR);
}



/* Text writing functions */
#define BUFSIZE 2048
#define ESC "\x01B"

#if (__DJGPP__)
#define vsnprintf(buf, BUFSIZE, format, va) vsprintf(buf, format, va)
#endif
#if (_WIN32)
#define vsnprintf _vsnprintf
#endif

/* Write a formatted string (like printf) with handling for special characters
   and color escapes.

   ESC f<x>	change foreground color to <x>
   ESC b<x>	change background color to <x>

   <x> is given as a single hexadecimal digit

   printc() will return the length of the printed string */

int printc(char *format, ...)
{
	unsigned char buf[BUFSIZE], tmp;
	int length,i;
	va_list va;
	color c;

	/* Print into buffer */
	va_start(va, format);
	length = vsnprintf(buf, BUFSIZE, format, va);
	va_end(va);
	if (length == -1) return -1;

	/* Print to screen */
	for (i=0; i<length; i++)
	{
		if (buf[i] == 10)
			gotoxy(1, wherey()+1);
		else if ((buf[i] == 27) && (
			(buf[i+1] == 'f') ||
			(buf[i+1] == 'b')) )
		{
			tmp = toupper( buf[i+2] );

			if ((tmp >= 48) && (tmp <= 48+9))
				c = tmp - 48;
			else if ((tmp >= 65) && (tmp <= 70))
				c = tmp - 65 + 10;
			else c = LightGray;

			if (buf[i+1] == 'f')
				fgcolor(c);
			else bgcolor(c);

			i += 2;
		}
		else
#if (USE_NCURSES) && (!__DJGPP__)
			writech(_c((chtype)(buf[i])));
#else
			writech(buf[i]);
#endif
	}

	return length;
}



/* Write a formatted string (works like printf) at x,y */
int writexy (int x, int y, char *format, ...)
{
	char buf[BUFSIZE];
	int length;
	va_list va;

	/* Move cursor */
	gotoxy(x, y);

	/* Print into buffer */
	va_start(va, format);
	length = vsnprintf(buf, BUFSIZE, format, va);
	va_end(va);
	if (length == -1) return -1;

	/* Spit it out */
	printc(buf);

	/* Done */
	return length;
}



/* Centerprint a formatted string (works like printf) at line y */
int center (int y, char *format, ...)
{
	char buf[BUFSIZE];
	int length, tmp_len, i;
	va_list va;

	/* Print into buffer */
	va_start(va, format);
	tmp_len = length = vsnprintf(buf, BUFSIZE, format, va);
	va_end(va);
	if (length == -1) return -1;

	/* Account for color escape codes */
	for (i=0; i<length; i++)
	if ((buf[i] == 27) && (
		(buf[i+1] == 'f') ||
		(buf[i+1] == 'b') ))
	{
		tmp_len -= 3;
		i += 2;
	}

	/* Move cursor */
	gotoxy((CON_COLS - tmp_len)/2, y);

	/* Spit it out */
	printc(buf);

	/* Done */
	return length;
}


 
/* Fill screen with color f,b and character c */
void fillscreen(color f, color b, chtype c)
{
	int x, y;

	/* Set the color we're going to use */
	if (has_color())
		setcolor(f,b);

	/* Fill the screen using repeated writech() calls */
	for (y=0; y<CON_ROWS; y++)
	{
		gotoxy(1,y+1);
		for (x=0; x<CON_COLS; x++) writech(c);
	}

	gotoxy(1,1);
}



#define MAXLINES 1000
/* Just a textviewer. -- Volker */
void textviewer (int x, int y, int w, int h, char *ttbuf)
{
	int tpos, posx, posy, scrollx, scrolly, numlines;
	int last_nl, last_space;
	int result = 101;
	int lines[MAXLINES];
	char *tbuf;
	key c;
	cursor cs = get_cursor();

	set_cursor(none);

	tpos = posx = posy = scrollx = scrolly = numlines = 0;
	last_nl = last_space = 0;

	tbuf = (char*) xmalloc(strlen(ttbuf)+1);
	strcpy(tbuf, ttbuf);
	tbuf[strlen(tbuf)] = 32;

	/* Wordwrapping */
	for (tpos=0; tpos<strlen(tbuf); tpos++)
	{
                switch (tbuf[tpos])
		{
                case 10:
			if (last_nl-2 > w)
			{
				tbuf[tpos-last_space+1] = 10;
				last_nl = last_space;
			}
			last_space = 1;
			last_nl = 1;

		case 32:
			if (last_nl-2 > w)
			{
				tbuf[tpos-last_space+1] = 10;
				last_nl = last_space;
			}
			last_space = 1;

		default:
			last_nl++;
			last_space++;
		}
	}

	/* Save the position of each line */
	lines[0] = 0;
	for (tpos=0; tpos<strlen(tbuf); tpos++)
	{
		if (tbuf[tpos] == 10)
		{
			if (++numlines < MAXLINES)
			{
				lines[numlines] = tpos + 1;
			}
		}
	}
	lines[++numlines] = strlen(tbuf);

	drawbox(x, y, x+w+1, y+h+1);

	do
	{
		/* Fill the box */
		for (posy=0; posy<h; posy++)
		{
			for (posx=0; posx<w; posx++)
			{
				gotoxy(x+posx+1, y+posy+1);
				tpos = lines[posy+scrolly] + posx + scrollx;
				if ((tpos<strlen(tbuf)) &&
				    (tpos < lines[posy+scrolly+1]))
				{
					if (tbuf[tpos] == 10)
					{
						for (; posx<w; posx++)
							printw(" ");
					}
					else
					{
						printw("%c", tbuf[tpos]);
					}
				}
				else
					printw(" ");
			}
		}

		update();
		c = readkey();

		/* Handle input */
		switch (c)
		{
		// filter
		case KEY_ENTER:
			result = 0;
			break;

		case KEY_ESC:
			result = 27;
			break;

		case KEY_UP:
			if (scrolly) scrolly--;
			break;

		case KEY_DOWN:
			if (scrolly + h < numlines) scrolly++;
			break;

		case KEY_LEFT:
			if (scrollx) scrollx--;
			break;

		case KEY_RIGHT:
			if (scrollx <= 500) scrollx++;
			break;

		case KEY_PGUP:
			if (scrolly >= h) scrolly -= h;
			else scrolly = 0;
			break;

		case KEY_PGDOWN:
			if (scrolly+2*h <= numlines) scrolly += h;
			else if (numlines > h) scrolly = numlines-h;
			break;

		case KEY_HOME:
			scrollx = 0;
			scrolly = 0;
			break;

		case KEY_END:
			scrollx = 0;
			if (numlines > h) scrolly = numlines-h;
			break;

		default:
			beep();
			break;
		}
	}
	while (result == 101);

	set_cursor(cs);

	xfree(tbuf);

}

