
/* Terminality - a portable terminal handling library
 * Copyright (C) 1998-2002, Emil Mikulic.
 * This is LGPL - look at COPYING.LIB
 */

/* Project:     Terminality/GUI
 * File:	menu.cpp
 * Author:      Emil Mikulic, Michal Safranek
 * Description: Implement the menuitem and menu classes
 */

/* Define if you want infinite menu loop (last->first && last<-first) */
#define INFINITE_LOOP

#include <gui.h>

const char menu_rcsid[] =
"$Id: menu.cpp,v 1.14 2002/07/26 01:39:40 darkmoon Exp $";



// Initialise menuItem
menuItem::menuItem()
{
	name = NULL;
	id = -1;
};



// Initialise menuItem
menuItem::menuItem(char *n, int i)
{
	name = (char *)xmalloc(strlen(n)+1);
	strcpy(name, n);
	id = i;
};



// Destroy menuItem
menuItem::~menuItem()
{
	xfree(name);
};



// Initialise menu list
menu::menu(menuItem *i, ...)
{
	va_list va;
	menuItem *curr;

	visible = fixed_colors = false;
	width = height = 0;
	current = 0;
	item = NULL;

	// colors
	SCROLL = COLOR_MENU_SCROLL;
	FRAME = COLOR_MENU_FRAME;
	CORNER = COLOR_MENU_CORNER;
	ITEM = COLOR_MENU_ITEM;
	FRAME_BG = COLOR_MENU_FRAMEBG;
	ITEM_BG = COLOR_MENU_ITEMBG;
	SELECTED_ITEM = COLOR_MENU_SITEM;
	SELECTED_ITEM_BG = COLOR_MENU_SITEMBG;

	// start the va_list
	va_start(va, i);
	curr = i;

	// as long as the item isn't a NULL
	while (curr) {
		// add to queue
		if (!item)
			item = tn_list_new((void*)curr);
		else
			tn_list_add(item, (void*)curr);

		// check against width
		if (strlen(curr->name) > (unsigned) width)
			width = strlen(curr->name);

		// get next item
		curr = va_arg(va, menuItem*);
	}
	va_end(va);

	// get list size
	count = tn_list_size(item);

	// fix menu height
	height = count;
	if (height > CON_ROWS-2) 
		height = CON_ROWS-2;
	scroll = 0;

	// make up x/y coords
	x = (CON_COLS-width-2) / 2;
	y = (CON_ROWS-height-2) / 2;
	if (!y) y=1;
	reg_id = register_add(Menu, this);
}



// destroy menu list
menu::~menu()
{
	// first kill all menuItems
	menuItem *tmp;

	for (int i=0; i<count; i++) {
		tmp = (menuItem*)tn_list_getdata(item, i);
		delete tmp;
	}
	if (item) {
		tn_list_free(item);
		tn_list_kill(item);
	}
	register_del(reg_id);
}



// draw menu
void menu::draw(void) const
{
	int i;

	if (!visible) return;
	bgcolor(FRAME_BG);

	// Go to upper left corner
	gotoxy(x,y);
	fgcolor(CORNER);
	writech(C_UL);	// upper-left

	// Draw top line
	fgcolor(FRAME);
	for (i=0; i<width; i++)
		writech(HLINE);
	fgcolor(CORNER);
	writech(C_UR); // upper-right

	// Draw vertical lines
	fgcolor(FRAME);
	for (i=1; i<=height; i++) {
		gotoxy(x,y+i);
		writech(VLINE);
		gotoxy(x+width+1,y+i);
		writech(VLINE);
	}

	// Go to lower left corner
	fgcolor(CORNER);
	gotoxy(x,y+height+1);
	writech(C_LL);	// lower-left

	// Draw bottom line
	fgcolor(FRAME);	
	for (i=0; i<width; i++)
		writech(HLINE);
	fgcolor(CORNER);
	writech(C_LR); // lower-right

	// Write out items
	if (item) {
		setcolor(ITEM,ITEM_BG);
		for (i=0; i<height && i+scroll<height; i++) {
			// Go to the Right Place(tm)
			gotoxy(x+1, y+1+i);

			// Set selected colour if current item
			if (i+scroll == current)
				setcolor(SELECTED_ITEM,SELECTED_ITEM_BG);

			// Check for separator
			menuItem *tmp;
			tmp = (menuItem*)tn_list_getdata(item, i+scroll);
			if (tmp->id == -1) {
				// safe enough
				for (int j=0; j<width; j++)
					writech(HLINE);
			} else {
				// Print item name
				printw(tmp->name);

				// Finish padding line with highlight bar
				for (int pad=strlen(tmp->name); pad<width;
				pad++)
					writech(' ');
			}

			// If current item then return normal colors
			if (i+scroll == current)
				// Return to normal colors
				setcolor(ITEM,ITEM_BG);
		}
	}

	// Draw scrollers
	setcolor(SCROLL,FRAME_BG);

	// Scroll up
	if (scroll) {
		gotoxy(x+2, y);
		writech(ARROW_U);
	}

	// Scroll down
	if (scroll+height < count) {
		gotoxy(x+width-1, y+height+1);
		writech(ARROW_D);
	}
}



// get choice and return menuitem id
// returns -1 on abort (ESC)
int menu::getchoice(void)
{
	key k;
	menuItem *temp;
	cursor cs = get_cursor();
	bool old_vis = visible;

	visible = true;	
	set_cursor(none);

	while (1) {
		draw();	// draw menu
		update();
		k = readkey();  // read key
		k = keyhandler(k, Menu);
		k = handle_key(k, Menu, this);


		switch (k) { // handle keypress
		case KEY_NOTHING: break;
		case KEY_UP: 
#ifdef INFINITE_LOOP
			if (current) {
				current--;
				if (current<scroll) scroll--;
			} else {
				scroll = count - height;
				current = count - 1;
			}
#else
			if (current) current--;
			if (current<scroll) scroll--;
#endif
			break;

		case KEY_TAB:
		case KEY_DOWN: 
#ifdef INFINITE_LOOP
			if (current<count-1) {
				current++; 
				if (current>scroll+height-1) scroll++;
			} else {
				current = 0;
				scroll = 0;
			}
#else
			if (current<count-1) current++;
			if (current>scroll+height-1) scroll++;
#endif
			break;

		case KEY_HOME: 
			scroll = 0; 
			current = 0; 
			break;

		case KEY_END: 
			scroll = count - height; 
			current = count - 1; 
			break;

		case KEY_PGUP:
			current -= height-1; 
			if (current<0) current = 0;
			if (scroll>current) scroll = current;
			break;

		case KEY_PGDOWN:
			current += height-1;
			if (current >= count) current = count-1;
			if (current >= scroll+height)
				scroll += height;
			if (scroll+height > count)
				scroll = count-height;
			break;

		case KEY_ENTER:
			// can't accept an item with id -1
			temp = (menuItem*)tn_list_getdata(item, current);
			if (temp->id != -1){
				set_cursor(cs);
				visible = old_vis;
				return temp->id; 
			}
			break;

		case KEY_ESC:
			set_cursor(cs);
			visible = old_vis;
			return -1;

		default:
			beep();
		}
	} 
}



// set height
int menu::setheight(int h)
{
	if (h && h <= tn_list_size(item)) {
		height = h;
	}
	return height;
}



// Change scheme
void menu::change_scheme(void)
{
	if (fixed_colors) return;
	SCROLL = COLOR_MENU_SCROLL;
	FRAME = COLOR_MENU_FRAME;
	CORNER = COLOR_MENU_CORNER;
	ITEM = COLOR_MENU_ITEM;
	FRAME_BG = COLOR_MENU_FRAMEBG;
	ITEM_BG = COLOR_MENU_ITEMBG;
	SELECTED_ITEM = COLOR_MENU_SITEM;
	SELECTED_ITEM_BG = COLOR_MENU_SITEMBG;
}

